module Scanf

using Base.Ryu

export @scanf, @sscanf, scanf

const EOF = -1
const ARGNUM_TYPE = UInt16
const WIDTH_TYPE = UInt32
# whitespace characters in format strings
const WHITESPACE = b" \n\t\r\f\v"
# format escape character
const ESC = UInt8('%')
# format specifier to skip whitespace
const SKIP = UInt8('_')
# format indicator no-assign
const NOASSIGN = UInt8('*')
# format specifiers for character sets
const CSOPEN = UInt8('[')
const CSCLOSE = UInt8(']')
const CSNEG = UInt8('^')
const CSMINUS = UInt8('-')

# format specifier categories
const Ints = Union{Val{'d'},Val{'i'},Val{'u'},Val{'x'},Val{'X'},Val{'o'}}
const Floats = Val{'f'}
const Chars = Union{Val{'c'},Val{'C'}}
const Strings = Union{Val{'s'},Val{'S'}}
const Pointer = Val{'p'}
const HexBases = Union{Val{'x'},Val{'X'},Val{'a'},Val{'A'}}
const DecBases = Union{Val{'d'},Val{'u'}}
const OctBases = Val{'o'}
const IntBases = Val{'i'}
const PositionCounter = Val{'n'}
const Whitespace = Val{Char(SKIP)}
const CharSets = Union{Val{Char(CSOPEN)},Val{Char(CSNEG)}}
const EscapeChar = Val{Char(ESC)}

const DECIMAL = b"0123456789"
const OCTAL = b"01234567"
const HEXADECIMAL = b"0123456789ABCDEFabcdef"
abstract type AbstractSpec end

"""
    Spec{T}

Typed representation of a format specifier.

`T` is a `Val{'X'}`, where `X` is a valid format specifier character.

Special case `X == '_'` represents whitespace (not assigning)

See also `CharsetSpec` and `LiteralSpec`.
"""
struct Spec{T<:Val} <: AbstractSpec # T => %type => Val{'type'}
    assign::ARGNUM_TYPE
    width::WIDTH_TYPE
end

"""
    CharsetSpec{T,S}

Format specifier representing a character set (inclusive or exclusive)
"""
struct CharsetSpec{T<:Val,S} <: AbstractSpec
    assign::ARGNUM_TYPE
    width::WIDTH_TYPE
    set::S
end

"""
    LiteralSpec{S<:AbstractString}

Non-assigning Format specifier representing a literal matching string.
"""
struct LiteralSpec{S} <: AbstractSpec
    string::S
end

# replace double %% by single % 
function LiteralSpec(str::T) where {T<:AbstractString}
    if contains(str, "%%")
        str = replace(str, "%%" => "%")
        S = typeof(str)
    else
        S = T
    end
    LiteralSpec{S}(str)
end

"""
    Scanf.Format(format_str)

Create a C scanf-compatible format object that can be used for parse formatted texts.

The input `format_str` can include any valid format specifier character and modifiers.

A `Format` object can be passed to `scanf(str::String, f::Format, args...)` to parse a
formatted string, or `scanf(io::IO, f::Format, args...)` to parse the
formatted string directly from `io`.

For convenience, the `Scanf.format"..."` string macro form can be used for building
a `Scanf.Format` object at macro-expansion-time.

The `@scanf` macro converts a literal string into a `Scanf.Format` implicitly.
"""
struct Format{S,T}
    str::S # original full format string as CodeUnits
    formats::T # Tuple of Specs (including whitespace specs and literal specs)
end

"""
    Scanf.scanf(b::String, f::Scanf.Format, args...)
    Scanf.scanf([io::IO,] f::Scanf.Format, args...)

Apply a Scanf format object `f` to provided String
(1st method), or scan directly from an `io` object (2nd method). See [`@scanf`](@ref)
for more details on C `scanf` support.
`io` defaults to `stdin`.
The args determine the type and default value for the output data.

Return the number of assigned arguments, followed by the assigned output data.
"""
function scanf end

scanf(io::IO, f::Format, args...) = scanner(io, f, args...)
scanf(f::Format, args...) = scanf(stdin, f, args...)
scanf(str::AbstractString, f::Format, args...) = scanf(IOBuffer(str), f, args...)

"""
    @scanf([io:Union{IO,String}, ], "%Fmt", args::...)

Scan input stream or string of using C `scanf` style format specification string and assign values 
to arguments.

The format string (not a variable) is analyzed at macro expansion time.
Equivalent to `scanf(io, Scanf.format"%Fmt", args...)`.

# Examples
```jldoctest
r, a... = @scanf("%f%f%f%f", zeros(4)...)

julia> r, a, b = @scanf "23.4 text" "%f %2s" Float64 String
(2, 23.4, "te")
```
"""
macro scanf(arg, args...)
    n = length(args)
    if arg isa String && !(n > 0 && args[1] isa String)
        fmt = arg
        io = :stdin
    elseif n >= 1 && args[1] isa String
        fmt = args[1]
        io = arg
        args = Base.tail(args)
    else
        throw(ArgumentError("invalid macro arguments: @scanf $arg $(join(args, ' '))"))
    end
    f = Format(unescape_string(fmt))
    return esc(:($Scanf.scanf($io, $f, $(args...))))
end

"""
    Scanf.format"..."

Convenience string macro to call `Scanf.Format("...")
"""
macro format_str(str)
    str = unescape_string(str)
    esc(:(Scanf.Format($str)))
end

# Implementation details

# construct Format object by parsing the format string
function Format(f::AbstractString)
    isempty(f) && throw(ArgumentError("empty format string"))
    bytes = codeunits(f)
    len = length(bytes)
    fmts = AbstractSpec[]
    pos, b = pushliteral!(fmts, f, bytes, 1)
    ac = 0 # count number of assigning specifiers

    while pos <= len || b == SKIP
        assign = true
        width = 0
        charset = nothing
        if b == ESC
            b = bytes[pos]
            pos += 1
            # positioned at start of first format str %
            # parse flags
            if b == NOASSIGN
                assign = false
                pos > len && throw(ArgumentError("incomplete format string: '$f'"))
                b = bytes[pos]
                pos += 1
            end
            # parse width
            while 0x00 <= b - UInt8('0') < 0x0a
                width = 10 * width + (b - UInt8('0'))
                b = bytes[pos]
                pos += 1
                pos > len && break
            end
            0 <= width <= typemax(WIDTH_TYPE) || throw(
                ArgumentError(
                    "invalid format string: \"$f\", width not in 1..$(typemax(Int32))",
                ),
            )
            # type modifier characters (ignored)
            if b == UInt8('h') || b == UInt8('l')
                prev = b
                pos > len && throw(ArgumentError("incomplete format string: \"$f\""))
                b = bytes[pos]
                pos += 1
                if b == prev # cases "ll" and "hh"
                    pos > len && throw(ArgumentError("incomplete format string: \"$f\""))
                    b = bytes[pos]
                    pos += 1
                end
            elseif b in b"Ljqtz" # more type modifiers (ignored)
                b = bytes[pos]
                pos += 1
            end
            # parse conversion specifier
            if b in b"diouxXcspn" # uper case variants are invalid
                type = Val{lowercase(Char(b))}
            elseif b in b"eEfFgGaA"
                type = Val{Char('f')}
            elseif b == CSOPEN
                txt = "["
                start = pos
                xpos = findnext(isequal(CSCLOSE), bytes, start)
                if start <= len && bytes[start] == CSNEG
                    b = CSNEG
                    txt = "[^"
                    start += 1
                end
                xpos === nothing && throw(
                    ArgumentError(
                        "invalid format string \"$f\", after '$txt' a Char(CSCLOSE) is missing",
                    ),
                )
                if bytes[start] == CSCLOSE # first character after "[" or "[^"
                    xpos = findnext(isequal(CSCLOSE), bytes, start + 1)
                end
                (xpos === nothing || start >= xpos) && throw(
                    ArgumentError(
                        "invalid format string \"$f\", after '$txt]' a Char(CSCLOSE) is missing",
                    ),
                )
                charset = view(bytes, start:xpos-1)
                pos = xpos + 1
                type = Val{Char(b)}
            else
                throw(
                    ArgumentError(
                        "invalid format string: \"$f\", invalid conversion specifier: '$(Char(b))'",
                    ),
                )
            end
        else # format contains a WS character
            type = Whitespace
            assign = false
        end
        ac += assign
        ass = ifelse(assign, ac, 0)
        ass <= typemax(ARGNUM_TYPE) || throw(
            ArgumentError(
                "invalid format string: \"$f\", too many assignable conversion specifiers",
            ),
        )
        push!(fmts, make_spec(type, ass, width, charset))
        pos, b = pushliteral!(fmts, f, bytes, pos)
    end
    return Format(bytes, Tuple(fmts))
end

# consume bytes up to next % or whitespace and insert LiteralSpec 
@inline function pushliteral!(fmts, f, bytes, pos)
    len = length(bytes)
    start = pos
    b = 0x00
    j = -1
    while pos <= len
        b = bytes[pos]
        pos += 1
        if b == ESC
            pos > len && throw(ArgumentError("incomplete format string: \"$f\""))
            if bytes[pos] == ESC # "%%" will be removed in LiteralSpec
                pos += 1
            else
                j = pos - 2
                break
            end
        elseif b in WHITESPACE
            j = pos - 2
            while pos <= len && bytes[pos] in WHITESPACE
                pos += 1
            end
            b = SKIP
            break
        end
    end
    j = j < 0 ? pos - 1 : j
    start <= j && push!(fmts, LiteralSpec(SubString(f, start:j)))
    return pos, b
end

Format(f::Format) = f

# match literal strings
@inline function fmt(io, pos, arg, res, spec::LiteralSpec)
    str = spec.string
    len = sizeof(str)
    ix = 1
    while ix <= len && !eof(io)
        c = str[ix]
        peek(io, Char) != str[ix] && break
        n = ncodeunits(c)
        skip(io, ncodeunits(c))
        ix = nextind(str, ix)
    end
    l = ix - 1
    pos + l, l == len
end

# skip whitespace spec
@inline function fmt(io, pos, arg, res, spec::Spec{Whitespace})
    skip_ws(io, pos), true
end

# match character(s)
@inline function fmt(io, pos, arg, res, spec::Spec{T}) where {T<:Chars}
    assign, j = assignnr(spec)
    width = spec.width
    width = ifelse(width == 0, 1, width)
    eof(io) && return pos, false

    if assign
        S = outtype(arg[j])
        v = S <: Char ? arg[j] : Char[]
    end
    i = 0
    start = pos
    while !eof(io) && i < width
        r = peek(io, Char)
        isvalid(Char, r) || break
        if assign
            i += 1
            assignto!(v, res, j, r, i)
        end
        n = ncodeunits(r)
        skip(io, n)
        pos += n
    end
    if assign && v isa AbstractVector
        resize!(v, i)
        v != arg[j] && assignto!(S, res, j, v)
    end
    return pos, pos > start
end

# string spec
@inline function fmt(io, pos, arg, res, spec::Spec{T}) where {T<:Strings}
    pos = skip_ws(io, pos)
    assign, j = assignnr(spec)
    width = spec.width
    l = 0
    out = IOBuffer()
    while (width == 0 || l < width) && !eof(io)
        c = peek(io, Char)
        (isspace(c) || !isvalid(Char, c)) && break
        assign && write(out, c)
        n = ncodeunits(c)
        skip(io, n)
        pos += n
        l += 1
    end
    if assign && l > 0
        r = String(take!(out))
        S = outtype(arg[j])
        x = scanf_parse(S, r)
        assignto!(arg[j], res, j, x)
    end
    return pos, l > 0
end

# charset specs
@inline function fmt(io, pos, arg, res, spec::CharsetSpec)
    assign, j = assignnr(spec)
    width = spec.width
    width = ifelse(width == 0, typemax(Int), width)
    l = 0
    out = IOBuffer()
    while l < width && !eof(io)
        c = peek(io, Char)
        n = ncodeunits(c)
        check_set(c, spec) || break
        skip(io, n)
        write(out, c)
        l += n
    end
    if assign && l > 0
        r = String(take!(out))
        S = outtype(arg[j])
        x = scanf_parse(S, r)
        assignto!(arg[j], res, j, x)
    end
    pos + l, l > 0
end

# integer specs
@inline function fmt(io, pos, arg, res, spec::Spec{T}) where {T<:Ints}
    pos = skip_ws(io, pos)
    assign, j = assignnr(spec)
    width = spec.width
    width = ifelse(width == 0, typemax(Int), width)
    l = 0
    sig = false
    eof(io) && return pos, false
    out = IOBuffer()
    b = peek(io)
    negate = false
    if b in b"+-"
        l = writeskip(io, out, b, l)
        negate = b == UInt('-')
        sig = true
    end
    digits = DECIMAL
    base, digits = basespec(T)
    ndig = 0
    if base === nothing || base == 16
        b = peek(io)
        if b == UInt('0')
            ndig += 1
            l = writeskip(io, out, b, l)
            if !eof(io) && peek(io) in b"xX"
                digits = HEXADECIMAL
                base = 16
                skip(io, 1)
                l += 1
                ndig = 0
            elseif base === nothing
                digits = OCTAL
                base = 8
            end
        end
    end
    while l < width && !eof(io)
        b = peek(io)
        if b in digits
            ndig += 1
            l = writeskip(io, out, b, l)
        else
            break
        end
    end
    succ = ndig > 0
    if assign && succ
        S = inttype(outtype(arg[j]))
        r = String(take!(out))
        x = scanf_parse(S, r, base = base)
        assignto!(arg[j], res, j, x)
    end
    pos + l, succ
end

# pointer spec
@inline function fmt(io, pos, arg, res, spec::Spec{T}) where {T<:Pointer}
    pos = skip_ws(io, pos)
    assign, j = assignnr(spec)
    width = spec.width
    width = ifelse(width == 0, typemax(Int), width)
    rv = (pos, false)
    eof(io) && return rv
    digits = HEXADECIMAL
    out = IOBuffer()
    l = 0
    while l < width && !eof(io)
        b = peek(io)
        if l == 0 && b == UInt8('0')
            l = writeskip(io, out, b, l)
            if !eof(io) && ( b = peek(io)) in b"xX"
                l = writeskip(io, out, b, l)
            end
        elseif b in digits
            l = writeskip(io, out, b, l)
        else
            break
        end
    end
    succ = l > 2
    if assign && succ
        r = String(take!(out))
        S = UInt
        s = tryparse(S, r)
        if s === nothing
            s = typemax(S)
        end
        assignto!(arg[j], res, j, s)
    end
    pos + l, succ
end

# floating point spec
@inline function fmt(io, pos, arg, res, spec::Spec{T}) where {T<:Floats}
    pos = skip_ws(io, pos)
    assign, j = assignnr(spec)
    width = spec.width
    width = ifelse(width == 0, typemax(Int), width)
    digits = DECIMAL
    expch = b"eEfF"
    x_sign = 0x01
    x_sep = 0x02
    x_exp = 0x04
    x_base = 0x08
    x_inexp = 0x10
    x_nexp = 0x20
    x_mdigits = 0x40
    x_edigits = 0x80
    l = 0
    out = IOBuffer()
    status = x_sign | x_sep | x_base | x_edigits
    while l < width && !eof(io)
        b = peek(io)
        if status & x_base != 0 && b == UInt8('0')
            l = writeskip(io, out, b, l)
            if !eof(io) && (b = peek(io)) in b"xX"
                digits = HEXADECIMAL
                expch = b"pP"
                l = writeskip(io, out, b, l)
            else
                status |= x_mdigits
            end
            status = (status | x_exp | x_edigits) & ~(x_base | x_sign)
            continue
        elseif b in digits
            status = (status | x_exp ) & ~(x_base | x_sign)
            status |= (status & x_inexp != 0) ? x_edigits : x_mdigits
        elseif b == UInt8('-') && (status & x_sign) != 0
            status |= ifelse(status & x_base == 0, x_nexp, 0)
            status = status & ~x_sign
        elseif b == UInt8('+') && (status & x_sign) != 0
            status = status & ~x_sign
        elseif b == UInt8('.') && (status & x_sep) != 0
            status = status & ~(x_base | x_sign | x_sep)
        elseif b in expch && (status & x_exp) != 0
            status = (status & ~(x_base | x_exp | x_edigits | x_sep)) | x_sign | x_inexp
            digits = DECIMAL
        elseif b in b"iI"
            n = expect(io, out, b"INFINITY", 3)
            l = (n == 3 || n == 8) ? l + n : 0
            status |= x_mdigits | x_edigits
            break
        elseif b in b"nN"
            n = expect(io, out, b"NAN", 3)
            l = n == 3 ? l + n : 0
            status |= x_mdigits | x_edigits
            break
        else
            break
        end
        l = writeskip(io, out, b, l)
    end
    succ = l > 0 && (status & x_mdigits != 0) && (status & x_edigits != 0)
    if assign && succ
        r = String(take!(out))
        A = floattype(outtype(arg[j]))
        A <: Integer && (A = float(A))
        s = scanf_parse(A, r, negx = status & x_nexp != 0)
        assignto!(arg[j], res, j, s)
    end
    pos + l, succ
end

# helpers for matching nan and inf
@inline function expect(io::IO, out::IO, bytes::AbstractVector{UInt8}, n)
    l = 1
    while l <= length(bytes) && !eof(io)
        b = peek(io)
        if b == bytes[l] || b == bytes[l] + (UInt8('a') - UInt8('A'))
            l <= n && write(out, b)
            skip(io, 1)
            l += 1
        else
            break
        end
    end
    return l - 1
end

# position counter spec
@inline function fmt(io, pos, arg, res, spec::Spec{PositionCounter})
    assign, j = assignnr(spec)
    if assign
        assignto!(arg[j], res, j, pos - 1)
    end
    pos, true
end

# fmt helpers
@inline function scanf_parse(
    ::Type{S},
    r::String;
    negx::Bool = false,
    base = nothing,
) where {S<:AbstractString}
    r
end

@inline function scanf_parse(
    ::Type{A},
    r::AbstractString;
    base = nothing,
    negx::Bool = false,
) where {A<:AbstractFloat}
    s = tryparse(A, r)
    if s === nothing
        s = ifelse(negx, zero(A), typemax(A))
        r[1] == '-' && (s = -s)
    end
    s
end

@inline function scanf_parse(
    ::Type{S},
    r::String;
    negx::Bool = false,
    base = nothing,
) where {S<:Integer}
    sig = r[1] in "+-"
    negate = r[1] == '-'
    if sig && S <: Unsigned
        r = SubString(r, 2:length(r))
    end
    x = tryparse(S, r, base = base)
    if x === nothing
        x = typemax(S)
        negate && S <: Signed && (x = typemin(S))
    elseif negate && S <: Unsigned
        x = -x
    end
    x
end

# skip WS characters in io stream
@inline function skip_ws(io, pos)
    while !eof(io)
        b = peek(io)
        n = _ncodeunits(b)
        if n == 1
            if b in WHITESPACE
                skip(io, 1)
                pos += 1
            else
                break
            end
        else
            c = peek(io, Char)
            if isspace(c)
                skip(io, n)
                pos += n
            else
                break
            end
        end
    end
    pos
end

# consume 1 byte
@inline function writeskip(io, out, b, l)
    write(out, b)
    skip(io, 1)
    l + 1
end

# call the format specifiers aligned with the IO stream
@inline function scanner(
    io::IO,
    f::Format,
    args::Union{
        Type,
        Ptr,
        Real,
        AbstractChar,
        AbstractString,
        AbstractVector{<:AbstractChar}
    }...,
)
    n = length(args)
    m = countspecs(f)
    n == m || argmismatch(m, n)
    res = Any[valuefor(x) for x in args]
    eof(io) && return EOF, res...
    pos = 1

    # for each format, scan arg and next substring
    # unroll up to 8 formats
    formats = f.formats
    N = length(formats)
    j = 0
    UNROLL_UPTO = 8
    Base.@nexprs 8 i -> begin
        if N >= i
            fi = formats[i]
            pos, succ = fmt(io, pos, args, res, fi)
            succ || @goto BREAK
            j += assign(fi)
        end
    end
    if N > UNROLL_UPTO
        for i = UNROLL_UPTO+1:N
            fi = formats[i]
            pos, succ = fmt(io, pos, args, res, fi)
            succ || break
            j += assign(fi)
        end
    end
    @label BREAK
    return tuple(j, res...)
end

# utility functions

# default value for types
valuefor(v::AbstractVector{T}) where {T<:AbstractChar} = similar(v, 0)
valuefor(::Type{T}) where {T<:Union{Real,Ptr,Char}} = T(0)
valuefor(::Type{T}) where {T<:AbstractString} = T("")
valuefor(a::T) where {T<:Union{Real,AbstractChar,AbstractString,Ptr}} = a

# assign value to reference or vector element
function assignto!(arg::AbstractVector, res, j, r, i = 1)
    if i > length(arg)
        resize!(arg, i)
    end
    arg[i] = r
    res[j] = arg
end
function assignto!(::Type{Char}, res, j, r, i = 1)
    if i == 1
        res[j] = length(r) >= 1 ? r[1] : ' '
    end
end
function assignto!(::Type{T}, res, j, r, i = 1) where {T}
    res[j] = T(r)
end
function assignto!(a::Any, res, j, r, i = 1)
    res[j] = oftype(a, r)
end

# accessor functions
assignnr(::LiteralSpec) = false, 0
assignnr(spec::AbstractSpec) = !iszero(spec.assign), Int(spec.assign)
assign(spec::AbstractSpec) = assignnr(spec)[1]

# showing formats

# recreate the format specifier string from a typed Spec
Base.string(f::Spec{Val{T}}; modifier::String = "") where {T} =
    string(Char(ESC), !assign(f) ? "*" : "", f.width > 0 ? f.width : "", modifier, T)
Base.string(f::CharsetSpec{Val{Char(CSOPEN)}}; modifier::String = "") = string(
    Char(ESC),
    !assign(f) ? "*" : "",
    f.width > 0 ? f.width : "",
    modifier,
    Char(CSOPEN),
    showset(f.set),
    Char(CSCLOSE),
)
Base.string(f::CharsetSpec{Val{Char(CSNEG)}}; modifier::String = "") = string(
    Char(ESC),
    !assign(f) ? "*" : "",
    f.width > 0 ? f.width : "",
    modifier,
    Char(CSOPEN),
    Char(CSNEG),
    showset(f.set),
    Char(CSCLOSE),
)
Base.string(f::LiteralSpec) = string('"', f.string, '"')

Base.show(io::IO, f::AbstractSpec) = print(io, string(f))

# reconstruct internals of %[...]
showset(s::UnitRange) = "$(Char(first(s)))-$(Char(last(s)))"

function showset(t::Tuple{String,Vararg})
    s = t[1]
    m, s = special(Char(CSMINUS), s)
    b, s = special(Char(CSCLOSE), s)
    string(b, s, showset.(t[2:end])..., m)
end

showset(t::Tuple) = string(showset.(t)...)

showset(s::AbstractString) = isempty(s) ? "1-0" : s

function special(x::Char, s::AbstractString)
    if occursin(x, s)
        string(x), filter(!isequal(x), s)
    else
        "", s
    end
end

# internal representation of % and whitespace format specifiers
function make_spec(T, assign, width, cs)
    if cs === nothing || isempty(cs)
        Spec{T}(assign, width)
    else
        charset = make_charset(cs)
        CharsetSpec{T,typeof(charset)}(assign, width, charset)
    end
end

# details of internal representation of %[...] format specifiers
make_charset(cs::AbstractVector{UInt8}) = make_charset(String(cs))
function make_charset(cs::String)
    M = Char(CSMINUS)
    ch = String[]
    ra = Any[]
    start = 1
    stop = lastindex(cs)
    mind = findall(isequal(M), cs)
    a = start
    for i in mind
        if i > start && i < stop
            j = prevind(cs, i)
            k = nextind(cs, i)
            cj, ck = cs[j], cs[k]
            if cj < ck
                push!(ra, UInt32(cj):UInt32(ck))
                b = prevind(cs, j)
            elseif cj == ck
                b = j
            else
                b = prevind(cs, j)
            end
            push!(ch, SubString(cs, a:b))
            a = nextind(cs, k)
        end
    end
    push!(ch, SubString(cs, a:stop))
    str = String(unique(collect(string(ch...))))
    isempty(ra) ? str : isempty(str) ? (length(ra) == 1 ? ra[1] : (ra...,)) : (str, ra...)
end

# bases for integer specs
basespec(::Type{<:HexBases}) = 16, HEXADECIMAL
basespec(::Type{<:OctBases}) = 8, OCTAL
basespec(::Type{<:DecBases}) = 10, DECIMAL
basespec(::Type) = nothing, DECIMAL

# type conversion hints for integer specs
outtype(::T) where {T} = outtype(T)
outtype(::Type{<:AbstractVector{T}}) where {T} = T
outtype(::Type{T}) where {T} = T

inttype(::Type{<:Float64}) = Int64
inttype(::Type{<:Float32}) = Int32
inttype(::Type{<:Float16}) = Int16
inttype(::Type{<:BigFloat}) = BigInt
inttype(::Type{T}) where {T} = T
floattype(::Type{T}) where {T<:Integer} = float(T)
floattype(::Type{T}) where {T} = T

# check if character is in or not in charset
@inline check_set(c::Char, spec::CharsetSpec{Val{Char(CSOPEN)}}) = check_in(c, spec.set)
@inline check_set(c::Char, spec::CharsetSpec{Val{Char(CSNEG)}}) = check_!in(c, spec.set)
@inline check_in(c, set::AbstractString) = occursin(c, set)
@inline check_!in(c, set::AbstractString) = !occursin(c, set)
@inline check_in(c, set::UnitRange) = in(UInt32(c), set)
@inline check_!in(c, set::UnitRange) = !in(UInt32(c), set)
@inline check_in(c, set::Tuple) = any(x -> check_in(c, x), set)
@inline check_!in(c, set::Tuple) = all(x -> check_!in(c, x), set)

# count number of assigning format specifiers in format
countspecs(f::Format) = count(assign, f.formats)

import Base: ==
function ==(f::T, g::T) where {T<:Format}
    f.formats == g.formats
end

# length of utf8 encoding of character starting with this byte
@inline _ncodeunits(a::UInt8) = a < 0xc0 ? 1 : a < 0xe0 ? 2 : a < 0xf0 ? 3 : 4

# error
@noinline function argmismatch(a, b)
    throw(
        ArgumentError(
            "mismatch between # of format specifiers and provided args: $a != $b",
        ),
    )
end

end # module
